import numpy as np
import pandas as pd
import plotly.express as px
from bioturbation.soil import SoilProfile


def run(n_soil_layers, soil_layer_depth, initial_conc, earthworm_density, beta,
                 dt=86400, steady_state_tol=1e-12, max_iter=10000):
    """Perform the bioturabtion calculations"""

    # Set up the soil profile
    soil_profile = SoilProfile(n_soil_layers,
                               soil_layer_depth,        # Depth of the soil layers in the profile [m]
                               initial_conc,            # Initial concentration in each layer [kg/m3]
                               earthworm_density,       # Earthworm density for each layer [earthworms/m2]
                               beta)                    # Empirical parameter relating earthworm density to depth to mix [m3/s]

    # Fill data fill the initial concentrations (specified as input)
    data = [[layer.conc] for layer in soil_profile.soil_layers]
    data_t = [layer.conc for layer in soil_profile.soil_layers]
    t = 0

    # Perform bioturbation until concentration equal across the soil layers
    while t < max_iter + 1 and not equal(data_t, tol=steady_state_tol):
        soil_profile.bioturbation(dt)
        for l, layer in enumerate(soil_profile.soil_layers):
            data[l].append(layer.conc)
            data_t = [layer.conc for layer in soil_profile.soil_layers]
        t += 1
        # If not converged after 10000 time steps, exit and set values as None
        if t > max_iter:
            print(f'Steady state not reached after {max_iter} days.')

    return np.array(data)


def as_df(data):
    df = pd.DataFrame(np.swapaxes(data,0,1),
                      columns=[f'soil_layer_{i+1}' for i in np.arange(0,data.shape[0])])
    return df


def plot(data):
    fig = px.line(as_df(data), labels={'value': 'concentration', 'index': 'time'})
    fig.show()


def equal(lst, tol=1e-12):
    """Check if list (lst) is equal within the specified tolerance (tol)"""
    diff = abs(max(lst) - min(lst))
    return diff < tol


class SoilLayer:
    """Class representing a soil layer"""

    def __init__(self, depth, initial_conc, earthworm_density, beta):
        """Initialise this soil layer with given depth, concentration and earthworm density"""
        self.depth = depth                                  # m
        self.conc = initial_conc                            # kg/m3
        self.earthworm_density = earthworm_density          # individuals/m3
        self.beta = beta                                    # m4/s
        self.bioturbation_rate = (self.earthworm_density * self.beta) / self.depth


class SoilProfile:
    """Class representing a soil profile, consisting of a number of soil layers"""

    def __init__(self, n_soil_layers, soil_layer_depth, initial_conc, earthworm_density, beta):
        """Initialise the soil profile with given number of soil layers and parameters for those layers
            soil_layer_depth: [m]
            initial_conc: [kg/m3]
            earthworm_density: [individuals/m3]
            beta: Empirical fitting parameter [m4/s]"""
        # Create the soil layers in this profile
        self.soil_layers = []
        for i in range(0, n_soil_layers):
            self.soil_layers.append(SoilLayer(soil_layer_depth,
                                              initial_conc[i],
                                              earthworm_density[i],
                                              beta))

    def bioturbation(self, dt):
        """Perform bioturbation for the length of time t by mixing calculated depth of two layers together"""
        for l, layer in enumerate(self.soil_layers):
            fraction_of_layer_to_mix = layer.bioturbation_rate * dt
            # Don't do for the final layer
            if l < len(self.soil_layers) - 1:
                layer.conc = layer.conc + fraction_of_layer_to_mix * (self.soil_layers[l+1].conc - layer.conc)
                self.soil_layers[l+1].conc = self.soil_layers[l+1].conc \
                                             + fraction_of_layer_to_mix * (layer.conc - self.soil_layers[l+1].conc)
